@@ -9,7 +9,7 @@
  */
 
 /*---------------------------------------------------------------------------*\
-  $Id: poll-compat.h 182945 2009-03-18 14:24:27Z russell $
+  $Id$
 
   NAME
 
@@ -79,6 +79,8 @@
 #ifndef __AST_POLL_COMPAT_H
 #define __AST_POLL_COMPAT_H
 
+#include "asterisk/select.h"
+
 #ifndef AST_POLL_COMPAT
 
 #include <sys/poll.h>
@@ -114,4 +116,24 @@ int ast_internal_poll(struct pollfd *pArray, unsigned long n_fds, int timeout);
 
 #endif /* AST_POLL_COMPAT */
 
+/*!
+ * \brief Same as poll(2), except the time is specified in microseconds and
+ * the tv argument is modified to indicate the time remaining.
+ */
+int ast_poll2(struct pollfd *pArray, unsigned long n_fds, struct timeval *tv);
+
+/*!
+ * \brief Shortcut for conversion of FD_ISSET to poll(2)-based
+ */
+static inline int ast_poll_fd_index(struct pollfd *haystack, int nfds, int needle)
+{
+	int i;
+	for (i = 0; i < nfds; i++) {
+		if (haystack[i].fd == needle) {
+			return i;
+		}
+	}
+	return -1;
+}
+
 #endif /* __AST_POLL_COMPAT_H */
